/* Name: main.c
 * Project: SI570 USB controller
 * Author: Tom Baier DG8SAQ
 * based on ObDev's AVR USB driver by Christian Starkjohann
 * Creation Date: 2006-04-23
 * Tabsize: 4
 * Copyright: (c) 2006 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: Proprietary, free under certain conditions. See Documentation.
 */
// DON'T EDIT THIS FILE UNLESS YOU REALLY HAVE TO. EDIT interface.h INSTEAD!!!!!
/*
 * Fuse bit information:
 * Fuse high byte:
 * 0xdd = 0 1 0 1   1 1 0 1
 *        ^ ^ ^ ^   ^ \-+-/ 
 *        | | | |   |   +------ BODLEVEL 2..0 (brownout trigger level -> 2.7V)
 *        | | | |   +---------- EESAVE (preserve EEPROM on Chip Erase -> not preserved)
 *        | | | +-------------- WDTON (watchdog timer always on -> disable)
 *        | | +---------------- SPIEN (enable serial programming -> enabled)
 *        | +------------------ DWEN (debug wire enable)
 *        +-------------------- RSTDISBL (disable external reset -> disabled)
 *
 * Fuse low byte:
 * 0xe1 = 1 1 1 0   0 0 0 1
 *        ^ ^ \+/   \--+--/
 *        | |  |       +------- CKSEL 3..0 (clock selection -> HF PLL)
 *        | |  +--------------- SUT 1..0 (BOD enabled, fast rising power)
 *        | +------------------ CKOUT (clock output on CKOUT pin -> disabled)
 *        +-------------------- CKDIV8 (divide clock by 8 -> don't divide) 
 *
 *
 * !!!!!! Remember to set Fuse SELFPRGEN to allow bootloading
 */
#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/boot.h>
#include <util/delay.h>
#include <stdint.h>
#include "usbdrv/usbdrv.h"

// code & data segments for code modifiable by bootloader
// don't forget to tell the linker where to locate the segments by the linker option e.g.
// -Wl,-section-start=.rwf_seg=0x0A00
// -Wl,-section-start=.rw_seg=0x0A08
// -W1,-section-start=.rwd_seg=0x800140
// whereby 0x0A00 is the start byte address of the rewritable code segment,
// and 0x800140 is the start address of the data for the rewritable code. 
// The word address shown by the disassembler is 0x0500 for the above example. 
// The linker will raise an error if the 
// segment addresses overlap with already assigned memory space.
// !!!!!! Remember to set Fuse SELFPRGEN to allow bootloading
#define REWRITABLE __attribute__((section(".rw_seg"))) 
//#define REWRITABLE  // uncomment to check code size
#define FUNCTABLE __attribute__((section(".rwf_seg"))) 
//#define FUNCTABLE  // uncomment to check code size
#define STRINGS __attribute__((section(".rws_seg"))) 
//#define STRINGS  // uncomment to check code size
#define USERDATA __attribute__((section(".rwd_seg")))
//#define USERDATA   // uncomment to check data size

#define UNPROTECTED ((unsigned int)jump_table)	//start byte address of bootloadable flash
#include "bootloader.h"


char command;
unsigned char bootcmd;
unsigned int  bootpage;
unsigned char page[64];
unsigned char pagepointer=0;


uint8_t mcusr_mirror __attribute__((section (".noinit")));
void get_mcusr(void) \
      __attribute__((naked)) \
      __attribute__((section(".init3")));
    void get_mcusr(void)
    {
      mcusr_mirror = MCUSR;
      MCUSR = 0;
      wdt_disable();
    }

/* ------------------------------------------------------------------------- */
/* ------------------------ interface to USB driver ------------------------ */
/* ------------------------------------------------------------------------- */


// Use this method to access re-flashable code if AVR silicon does NOT support a JMP instruction 
#define pt2FunctionWrite(data,len) ((uchar(*)(uchar *data, uchar len)) pgm_read_word(&jump_table[0]) )(data,len)
#define pt2FunctionSetup(data)     ((uchar(*)(uchar data[8] ))         pgm_read_word(&jump_table[1]) )(data)
#define pt2TW_Action               ((void (*)(void))                   pgm_read_word(&jump_table[2]) )
#define pt2Timer0_Overflow         ((void (*)(void))                   pgm_read_word(&jump_table[3]) )
#define pt2InitVars                ((void (*)(void))                   pgm_read_word(&jump_table[4]) )

#include "interface.h"	//user interface to USB driver

//need to use &jump_table[0] [1] [2], etc.
uint16_t jump_table[] FUNCTABLE = {
	(uint16_t)(&FunctionWrite),
	(uint16_t)(&FunctionSetup),
	(uint16_t)(&TW_Action),
	(uint16_t)(&Timer0_Overflow),
	(uint16_t)(&InitVars)
	};




/*------------------------------------------------------------------------------*/
/*------------------------usbFunctionSetup(data)--------------------------------*/
/*	Called by USB interface code when a control message has been                */
/*    received.                                                                 */
/*    Returns: 0xFF if 'user supplied function'- in our case usbFunctionWrite() */
/*      - should be called.                                                     */
/*      Otherwise, number of bytes to return over USB link. Corresponding data  */
/*      is put into a buffer pointed to by usbMsgPtr                            */
/*                                                                              */
/*  (See USBDRV.H for details)                                                  */
/*  Note that most command processing is done here BUT commands which need      */
/*  access to the input data buffer need to be implemented in usbFunctionWrite  */
/*------------------------------------------------------------------------------*/

USB_PUBLIC uchar usbFunctionSetup(uchar data[8])
{
usbRequest_t *rq = (void *)data;
static uchar    replyBuf[SPM_PAGESIZE];
unsigned char i;
    usbMsgPtr = replyBuf;
	if(rq->bRequest == 0){       		// ECHO value
        replyBuf[0] = data[2];			// rq->bRequest identical data[1]!
        return 1;
    }
	if(rq->bRequest == 0x10){				// EEPROM write byte value=address, index=data
	    eeprom_write_byte((uint8_t *)(data[2]+256*data[3]), data[4]);  
        return 0;
    }
	if(rq->bRequest == 0x11){       		// EEPROM read byte "value"=address
        replyBuf[0] = 
		   eeprom_read_byte((uint8_t *)(data[2]+256*data[3]));
        return 1;
	}
	if(rq->bRequest == 0xEB){       		// ECHO value, diagnostic command
        replyBuf[0] = data[2];				// rq->bRequest identical data[1]!
        return 1;
    }
    if(rq->bRequest == 0xEC){ 				// EEPROM write byte to safety flag value=data
	    eeprom_write_byte((uint8_t *)SAFETY_FLAG, data[2]); 
        return 0;
    }
	if((rq->bRequest == 0xED) |  (rq->bRequest == 0xF)){       // Reset by Watchdog
        for (;;){};
    }
	if(rq->bRequest == 0xFE){     
	    GIMSK &= ~(1<<PCIE); 	    		// disable PCINT interrupt in the general interrupt mask
		TIMSK=TIMSK & ~(1<<TOIE0);			// disable Timer0 overflow interrupt
    }
	if(rq->bRequest == 0xEE)				// Reflash using bootloader		
	{    
		command = 0xEE;						// bootload or read flash command
		bootcmd=data[2];					// lower byte of value
	    bootpage=data[4]|(data[5]<<8);		// index=flash byte address
		if (bootcmd==0xEE)
		{
			if (bootpage<UNPROTECTED)
		  	{
				replyBuf[0]=0xAA;			// notify forbidden
			} else {
				boot_program_page (bootpage, page);	//selfflash
				replyBuf[0]=SPM_PAGESIZE;	// return flashed block size
			}
			return(1);	
		}
		if (bootcmd>0xEE)
			return 0xff;					// translate to USB function write to transfer block data for flashing
		for (i=0;i<SPM_PAGESIZE;i++)
			replyBuf[i]=pgm_read_byte (bootpage+i);	// read flash page
		return SPM_PAGESIZE;    						
	}

	if (rq->bRequest == 0xEF)				// Reply with Flashable area start address
	{
		((int16_t*)replyBuf)[0] = (int16_t)&jump_table;
		((int16_t*)replyBuf)[1] = (int16_t)FLASHEND+1;
		return 4;							// 2x2 bytes returned
	}
	return pt2FunctionSetup(data);  		// Everything else is a user command
}


/*------------------------------------------------------------------------------*/
/*------------------------usbFunctionWrite(data, len)---------------------------*/
/*	Called by USB interface code when a packet has been received and            */
/*  usbFunctionSetup returns 0xFF                                               */
/*    Returns: 0xFF if an error has occurred                                    */
/*             0 if more data is expected                                       */
/*             non-zero if all data received and processed                      */
/*                                                                              */
/*  (See USBDRV.H for details)                                                  */
/*------------------------------------------------------------------------------*/


USB_PUBLIC uchar usbFunctionWrite(uchar *data, uchar len) //sends len bytes to SI570
{
int i;
if (command != 0xEE) {return pt2FunctionWrite(data, len);}
for (i=0;i<len;i++){					//save data subblock
	page[pagepointer+i]=data[i];
	}
pagepointer=pagepointer+len;
if (pagepointer>=SPM_PAGESIZE){
	pagepointer=0;
	return 1;
	}
return 0;
}

/* ------------------------------------------------------------------------- */
/* ------------------------ Oscillator Calibration ------------------------- */
/* ------------------------------------------------------------------------- */

/* Calibrate the RC oscillator to 8.25 MHz. The core clock of 16.5 MHz is
 * derived from the 66 MHz peripheral clock by dividing. Our timing reference
 * is the Start Of Frame signal (a single SE0 bit) available immediately after
 * a USB RESET. We first do a binary search for the OSCCAL value and then
 * optimize this value with a neighboorhod search.
 * This algorithm may also be used to calibrate the RC oscillator directly to
 * 12 MHz (no PLL involved, can therefore be used on almost ALL AVRs), but this
 * is wide outside the spec for the OSCCAL value and the required precision for
 * the 12 MHz clock! Use the RC oscillator calibrated to 12 MHz for
 * experimental purposes only!
 */
static void calibrateOscillator(void)
{
uchar       step = 128;
uchar       trialValue = 0, optimumValue;
int         x, optimumDev, targetValue = (unsigned)(1499 * (double)F_CPU / 10.5e6 + 0.5);

    /* do a binary search: */
    do{
        OSCCAL = trialValue + step;
        x = usbMeasureFrameLength();    /* proportional to current real frequency */
        if(x < targetValue)             /* frequency still too low */
            trialValue += step;
        step >>= 1;
    }while(step > 0);
    /* We have a precision of +/- 1 for optimum OSCCAL here */
    /* now do a neighborhood search for optimum value */
    optimumValue = trialValue;
    optimumDev = x; /* this is certainly far away from optimum */
    for(OSCCAL = trialValue - 1; OSCCAL <= trialValue + 1; OSCCAL++){
        x = usbMeasureFrameLength() - targetValue;
        if(x < 0)
            x = -x;
        if(x < optimumDev){
            optimumDev = x;
            optimumValue = OSCCAL;
        }
    }
    OSCCAL = optimumValue;
}
/*
Note: This calibration algorithm may try OSCCAL values of up to 192 even if
the optimum value is far below 192. It may therefore exceed the allowed clock
frequency of the CPU in low voltage designs!
You may replace this search algorithm with any other algorithm you like if
you have additional constraints such as a maximum CPU clock.
For version 5.x RC oscillators (those with a split range of 2x128 steps, e.g.
ATTiny25, ATTiny45, ATTiny85), it may be useful to search for the optimum in
both regions.
*/

void    usbEventResetReady(void)
{
    calibrateOscillator();
    eeprom_write_byte(0, OSCCAL);   /* store the calibrated value in EEPROM */
}

/* ------------------------------------------------------------------------- */
/* --------------------------------- main ---------------------------------- */
/* ------------------------------------------------------------------------- */

int main(void)
{
uchar   calibrationValue;
if (eeprom_read_byte((uint8_t *)SAFETY_FLAG)==0xFF)
  {
   eeprom_write_byte((uint8_t *)SAFETY_FLAG,0x00);
   pt2InitVars();
  }

//  usbConnect sequence
    calibrationValue = eeprom_read_byte(0); /* calibration value from last time */
    if(calibrationValue != 0xff){
        OSCCAL = calibrationValue;
    	}
    usbDeviceDisconnect();
    _delay_ms(300);  // 300 ms disconnect 
    usbDeviceConnect();
//  usbConnect sequence done

	wdt_enable(WDTO_1S );	//watchdog 1s
    usbInit();
    sei();

    for(;;)
		{    /* main event loop */
        wdt_reset();
        usbPoll();
    	}
    //return 0;
}
